__author__ = 'Kyle'

import math
import re

import matplotlib.pyplot as plt
import numpy as np


class Variable:
    """Represent a variable as a class."""

    def __init__(self, coef, exp):
        """Set coefficient and exponent."""
        self.coef = coef
        self.exp = exp

    def __str__(self):
        """Convert variable to a string.

         Returns a string in the form Ax^B.
         """
        if isinstance(self.coef, float):
            if self.coef.is_integer():
                self.coef = int(self.coef)
        if isinstance(self.exp, float):
            if self.exp.is_integer():
                self.exp = int(self.exp)
        if self.coef == 0:
            return "0"
        if self.coef == 1:
            if self.exp == 0:
                return "1"
            elif self.exp == 1:
                return "x"
            return "x^" + str(self.exp)
        if self.exp == 0:
            return str(round(self.coef, 2))
        elif self.exp == 1:
            return str(round(self.coef, 2)) + "x"
        return str(round(self.coef, 2)) + "x^" + str(self.exp)


class Function:
    """Represent a function as a class."""

    def __init__(self, *variables):
        """Set function to the list of variables."""
        self.function = variables

    def __str__(self):
        """Convert the function to a string

        Returns a string in the form Ax^B + Cx^D ...
        """
        function = ""
        for var in self.function:
            function += str(var) + " + "
        return function[:-3]


class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __str__(self):
        return "(" + str(self.x) + ", " + str(self.y) + ")"


def quadratic_formula(a, b, c):
    """Find the zeros of a quadratic function.

    Takes the coefficients of the variables and returns the zeros as a list.
    """
    result = []
    root = math.sqrt(b ** 2 - 4 * a * c)
    top1 = -b + root
    top2 = -b - root
    result.append(top1 / 2 * a)
    result.append(top2 / 2 * a)
    return result


def operation_from_text(text):
    """Finds math operation from text and returns the result of the operation.

    Takes a text containing a math operation and returns
     the result of the math operation or None if no result can be returned.
    """
    numbers = get_numbers(text)
    if not (len(numbers) >= 1 or in_text('graph', text) or in_text('plot', text) or in_text('trend', text) or in_text(
            'fibo', text)):
        return None
    if in_text('+', text) or in_text('plus', text):
        return numbers[0] + numbers[1]
    elif in_text('-', text) or in_text('minus', text):
        return numbers[0] - numbers[1]
    elif in_text('*', text) or in_text("mult", text) or in_text('times', text):
        return numbers[0] * numbers[1]
    elif in_text('/', text) or in_text('div', text):
        return numbers[0] / numbers[1]
    elif in_text('hyp', text):
        return math.hypot(numbers[0], numbers[1])
    elif in_text('pow', text):
        return numbers[0] ** numbers[1]
    elif in_text('square', text) and in_text('root', text):
        return math.sqrt(numbers[0])
    elif in_text('square', text):
        return numbers[0] ** 2
    elif in_text('natural', text) and in_text('log', text):
        return math.log(numbers[0])
    elif in_text('log', text) and in_text('base', text):
        return math.log(numbers[1], numbers[0])
    elif in_text('log', text):
        return math.log10(numbers[0])
    elif in_text('area', text) and in_text('circle', text):
        return math.pi * numbers[0] ** 2
    elif in_text('area', text):
        return numbers[0] * numbers[1]
    elif in_text('circle', text):
        return math.pi * numbers[0] * 2
    elif in_text('factorial', text) or in_text('!', text):
        return math.factorial(numbers[0])
    elif in_text('factor', text):
        return prime_factor(numbers[0])
    elif in_text('fibo', text):
        if len(numbers) > 0:
            return fibonacci(numbers[0])
        else:
            return fibonacci(100)
    return None


def in_text(s, text):
    """Finds if a string is in a text.

    Takes the string to look for in text, (s) and the text to find a string (text).
    Returns true if s is in text, False if it is not.
    """
    return s in text


def get_numbers(text):
    """Finds and extracts numbers from a text.

    Takes a text to find numbers and returns the numbers found in the text as an array.
    """
    nums = re.findall(r'([0-9]?\.?[0-9]+)', text)
    return list(map(lambda x: float(x), nums))


def create_variable(coefficient, exponent):
    """Create a representation of a variable in math.

    Takes a coefficient of a variable and an exponent and creates a representation of a variable with it.
    Returns a Variable object.
    """
    return Variable(coefficient, exponent)


def create_point(x, y):
    """Create a representation of a point.

    Takes an x and y position and creates a representation of a point with it.
    Returns a Point object.
    """
    return Point(x, y)


def x(point):
    """Returns the x value of a point."""
    return point.x


def y(point):
    """Returns the y value of a point."""
    return point.y


def create_function(*variables):
    """Create a representation of a function in math.

    Takes several variables and creates a representation of a function with them.
    Returns a Function object.
    """
    return Function(*variables)


def str_to_fn(fn):
    """Create a representation of a function in math.

    Takes a string and creates a representation of a function with it.
    The string is in this format: '7x^6 - 9x^4 + 7x^2 - x + 2'.
    Returns a Function object.
    """
    raw = fn.split()
    while '+' in raw:
        raw.remove('+')
    while '-' in raw:
        index = raw.index('-')
        raw[index + 1] = '-' + raw[index + 1]
        raw.remove('-')
    variables = []
    for variable in raw:
        exp = None
        coef = None
        try:
            coef = re.findall(r'(-?[0-9]\.?[0-9]?)?x', variable)[0]
        except IndexError:
            pass
        if coef == '':
            coef = 1
        try:
            exp = re.findall(r'x\^?(-?[0-9]\.?[0-9]?)?', variable)[0]
        except IndexError:
            pass
        if exp == '':
            exp = 1
        if coef is None and exp is None:
            try:
                coef = int(variable)
            except ValueError:
                pass
            exp = 0
        variables.append(create_variable(float(coef), float(exp)))
    return create_function(*variables)


def derivative(fn):
    """Find the derivative of a function.

    Takes a Function and returns the derivative of the function.
    """
    dfn = []
    for variable in fn.function:
        if variable.exp == 0:
            continue
        else:
            dfn.append(create_variable(variable.coef * variable.exp, variable.exp - 1))
    return create_function(*dfn)


def antiderivative(dfn):
    """Find the antiderivative of a function.

    Takes a Function and returns the antiderivative of the function without C.
    """
    fn = []
    for variable in dfn.function:
        fn.append(create_variable(variable.coef / (variable.exp + 1), variable.exp + 1))
    return create_function(*fn)


def critical_points(fn):
    """Find the critical points of a function

    Takes a Function and returns a dictionary of the critical points. The dictionary keys and values are:
    'max' -> point, 'min' -> point, 'inflection points' -> tuple of points.
    """
    values = {}
    dfn = derivative(fn)
    first_cp = zeros(dfn)
    first_criticals = []
    for x_val in first_cp:
        first_criticals.append(create_point(x_val, evaluate(fn, x_val)))
    values['max'] = max(first_criticals, key=y)
    values['min'] = max(first_criticals, key=x)
    d2fn = derivative(dfn)
    second_cp = zeros(d2fn)
    second_criticals = []
    for x_val in second_cp:
        second_criticals.append(create_point(x_val, evaluate(fn, x_val)))
    values['inflection points'] = tuple(second_criticals)
    return values


def evaluate(fn, x, precision=2):
    """Find the y value at a point, x.

    Takes a Function and an x value and returns the y value at the point, x.
    """
    total = 0
    for var in fn.function:
        total += var.coef * x ** var.exp
    return round(total, precision)


def zeros(fn):
    """Find the rational zeros of a function.

    Takes a Function as input and returns a list of rational zeros.
    """
    p = fn.function[-1]
    if p.exp != 0:
        p = 0
    else:
        p = p.coef
    q = fn.function[0].coef
    p_factors = prime_factor(p)
    q_factors = prime_factor(q)
    possible_zeros = []
    for numerator in p_factors:
        for denominator in q_factors:
            if numerator / denominator not in possible_zeros:
                possible_zeros.append(numerator / denominator)
            if -numerator / denominator not in possible_zeros:
                possible_zeros.append(-numerator / denominator)
    rational_zeros = []
    for pz in possible_zeros:
        if evaluate(fn, pz) == 0:
            rational_zeros.append(pz)
    return rational_zeros


def integrate(fn, start, stop):
    """Find the area under the curve of a function.

    Takes a Function, a starting point, and a stopping point and returns the area under the curve.
    """
    a_fn = antiderivative(fn)
    return evaluate(a_fn, stop) - evaluate(a_fn, start)


def slope(p1, p2):
    """Find the slope of a line using two points.

    Takes two Points and returns the slope as a float.
    """
    return round((p1.y - p2.y) / (p1.x - p2.x), 2)


def prime_factor(x):
    """Breaks down a number into its prime factors.

    Returns a list of factors for a number, x.
    """
    if x == 0:
        return [0]
    nums = [1]
    divisor = 2
    negative = x < 0
    if negative:
        x = abs(x)
    if x >= 0:
        while not x == 1:
            while x % divisor == 0:
                x /= divisor
                nums.append(divisor)
            divisor += 1
    if negative:
        neg_nums = []
        for num in nums:
            neg_nums.append(-num)
        for num in neg_nums:
            nums.append(num)
    return nums


def fibonacci(n):
    """Calculates the fibonacci numbers up to n.

    Returns a list of fibonacci numbers up to n.
    """
    result = []
    a, b = 0, 1
    while b < n:
        result.append(b)
        a, b = b, a + b
    return result


def plot_from_csv(filename, title, line=True, trend=True, degree=1, x_label=None, y_label=None):
    """Create and show a graph from a csv file.

    Get data from the file, filename, and create a graph with the title, title. The parameters include
    a line graph, line (True), a trend line, trend (True), the degree of the trend line, degree (1),
    the x_label (None, in csv), and y_label (None, in csv). This returns a Function for the trend line.
    """
    x = []
    y = []
    if not filename.endswith(".csv"):
        filename += ".csv"
    test = open(filename, "r")
    separated_file = test.read().split('\n')
    test.close()
    label = False
    for item in separated_file:
        x_and_y = item.split(',')
        if not x_and_y[0].startswith('-') and not x_and_y[1].startswith('-'):
            if not x_and_y[0].isdigit() and x_label is None:
                plt.xlabel(x_and_y[0])
                label = True
            if not x_and_y[1].isdigit() and y_label is None:
                plt.ylabel(x_and_y[1])
                label = True
                continue
        if len(x_and_y) > 1:
            x.append(int(x_and_y[0]))
            y.append(int(x_and_y[1]))

    plt.title(title)
    if not label:
        plt.xlabel(x_label)
        plt.ylabel(y_label)
    if trend:
        z = np.polyfit(x, y, deg=degree)
        x_new = np.linspace(min(x) - 10, max(x) + 10, 1000)
        y_new = np.polyval(z, x_new)
        plt.plot(x_new, y_new, "r--")
    if line:
        plt.plot(x, y)
    else:
        plt.scatter(x, y)
    plt.show()
    if trend:
        if degree == 1:
            a = create_variable(z[0], 1)
            b = create_variable(z[1], 0)
            return create_function(a, b)
        elif degree == 2:
            a = create_variable(z[0], 2)
            b = create_variable(z[1], 1)
            c = create_variable(z[2], 0)
            return create_function(a, b, c)
        elif degree == 3:
            a = create_variable(z[0], 3)
            b = create_variable(z[1], 2)
            c = create_variable(z[2], 1)
            d = create_variable(z[3], 0)
            return create_function(a, b, c, d)


def trend_line(filename):
    """ Find the trend line of a csv file.

    Returns a Function for the trend line of the csv file, filename.
    """
    x = []
    y = []
    if not filename.endswith(".csv"):
        filename += ".csv"
    test = open(filename, "r")
    separated_file = test.read().split('\n')
    test.close()
    for item in separated_file:
        x_and_y = item.split(',')
        if len(x_and_y) > 1:
            x.append(int(x_and_y[0]))
            y.append(int(x_and_y[1]))
    z = np.polyfit(x, y, deg=1)
    a = create_variable(z[0], 1)
    b = create_variable(z[1], 0)
    return create_function(a, b)


def plot(fn, start=-10, stop=10):
    """ Plot a function.

    Plots the Function, fn, from start (-10) to stop (10).
    """
    x = np.arange(start, stop + 1)
    y = [evaluate(fn, result) for result in x]
    plt.plot(x, y)
    plt.show()

